'''
Date: 2024-05-06 17:19:43
LastEditors: BHM-Bob 2262029386@qq.com
LastEditTime: 2024-10-31 16:59:27
Description: 
'''

__title__ = "lazydock"
__description__ = "Lazy Dock"
__url__ = "https://github.com/BHM-Bob/LazyDock"
__version__ = "0.5.0"
__build__ = "20241031_0"
__author__ = "BHM-Bob G"
__author_email__ = "bhmfly@foxmail.com"
__license__ = "MIT"